 /*    This file is part of the Gudhi Library. The Gudhi library
  *    (Geometric Understanding in Higher Dimensions) is a generic C++
  *    library for computational topology.
  *
  *    Author(s):       David Salinas
  *
  *    Copyright (C) 2014 Inria
  *
  *    This program is free software: you can redistribute it and/or modify
  *    it under the terms of the GNU General Public License as published by
  *    the Free Software Foundation, either version 3 of the License, or
  *    (at your option) any later version.
  *
  *    This program is distributed in the hope that it will be useful,
  *    but WITHOUT ANY WARRANTY; without even the implied warranty of
  *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *    GNU General Public License for more details.
  *
  *    You should have received a copy of the GNU General Public License
  *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */

#ifndef SKELETON_BLOCKER_ITERATORS_SKELETON_BLOCKERS_ITERATORS_H_
#define SKELETON_BLOCKER_ITERATORS_SKELETON_BLOCKERS_ITERATORS_H_

#include <gudhi/Skeleton_blocker/iterators/Skeleton_blockers_vertices_iterators.h>
#include <gudhi/Skeleton_blocker/iterators/Skeleton_blockers_edges_iterators.h>
#include <gudhi/Skeleton_blocker/iterators/Skeleton_blockers_blockers_iterators.h>
#include <gudhi/Skeleton_blocker/iterators/Skeleton_blockers_triangles_iterators.h>
#include <gudhi/Skeleton_blocker/iterators/Skeleton_blockers_simplices_iterators.h>

#endif  // SKELETON_BLOCKER_ITERATORS_SKELETON_BLOCKERS_ITERATORS_H_
