



if not exist %LIBRARY_INC%\gudhi\Simplex_tree.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -c "import gudhi; diag1 = [[2.7, 3.7], [9.6, 14.0], [34.2, 34.974], [3.0, float('Inf')]]; diag2 = [[2.8, 4.45], [9.5, 14.1], [3.2, float('Inf')]]; assert gudhi.bottleneck_distance(diag1, diag2) == 0.75"
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
